@@ -48,6 +48,8 @@ static void timer_destroy_cb(pl_gpu gpu, pl_timer timer)
     pl_free(timer);
 }
 
+VK_CB_FUNC_DEF(timer_destroy_cb);
+
 static pl_timer vk_timer_create(pl_gpu gpu)
 {
     struct pl_vk *p = PL_PRIV(gpu);
@@ -72,15 +74,7 @@ error:
 
 static void vk_timer_destroy(pl_gpu gpu, pl_timer timer)
 {
-    struct pl_vk *p = PL_PRIV(gpu);
-    struct vk_ctx *vk = p->vk;
-    pl_mutex_lock(&p->recording);
-    if (p->cmd) {
-        vk_cmd_callback(p->cmd, (vk_cb) timer_destroy_cb, gpu, timer);
-    } else {
-        vk_dev_callback(vk, (vk_cb) timer_destroy_cb, gpu, timer);
-    }
-    pl_mutex_unlock(&p->recording);
+    vk_gpu_idle_callback(gpu, VK_CB_FUNC(timer_destroy_cb), gpu, timer);
 }
 
 static uint64_t vk_timer_query(pl_gpu gpu, pl_timer timer)
@@ -104,7 +98,7 @@ static uint64_t vk_timer_query(pl_gpu gpu, pl_timer timer)
     switch (res) {
     case VK_SUCCESS:
         timer->index_read = (timer->index_read + 2) % QUERY_POOL_SIZE;
-        return (ts[1] - ts[0]) * vk->limits.timestampPeriod;
+        return (ts[1] - ts[0]) * vk->props.limits.timestampPeriod;
     case VK_NOT_READY:
         return 0;
     default:
@@ -136,16 +130,12 @@ static void timer_begin(pl_gpu gpu, struct vk_cmd *cmd, pl_timer timer)
     if (cmd->pool->props.queueFlags & reset_flags) {
         // Use direct command buffer resets
         vk->CmdResetQueryPool(cmd->buf, timer->qpool, timer->index_write, 2);
-    } else if (p->host_query_reset) {
-        // Use host query resets
-        vk->ResetQueryPoolEXT(vk->dev, timer->qpool, timer->index_write, 2);
     } else {
-        PL_TRACE(gpu, "QF %d supports no mechanism for resetting queries",
-                 cmd->pool->qf);
-        return;
+        // Use host query reset
+        vk->ResetQueryPool(vk->dev, timer->qpool, timer->index_write, 2);
     }
 
-    vk->CmdWriteTimestamp(cmd->buf, VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
+    vk->CmdWriteTimestamp(cmd->buf, VK_PIPELINE_STAGE_ALL_COMMANDS_BIT,
                           timer->qpool, timer->index_write);
 
     p->cmd_timer = timer;
@@ -174,8 +164,8 @@ struct vk_cmd *_begin_cmd(pl_gpu gpu, enum queue_type type, const char *label,
     }
 
     if (!p->cmd || p->cmd->pool != pool) {
-        vk_cmd_submit(vk, &p->cmd);
-        p->cmd = vk_cmd_begin(vk, pool, label);
+        vk_cmd_submit(&p->cmd);
+        p->cmd = vk_cmd_begin(pool, label);
         if (!p->cmd) {
             pl_mutex_unlock(&p->recording);
             return NULL;
@@ -208,7 +198,7 @@ bool _end_cmd(pl_gpu gpu, struct vk_cmd **pcmd, bool submit)
     if (!pcmd) {
         if (submit) {
             pl_mutex_lock(&p->recording);
-            ret = vk_cmd_submit(p->vk, &p->cmd);
+            ret = vk_cmd_submit(&p->cmd);
             pl_mutex_unlock(&p->recording);
         }
         return ret;
@@ -219,11 +209,11 @@ bool _end_cmd(pl_gpu gpu, struct vk_cmd **pcmd, bool submit)
 
     if (p->cmd_timer) {
         pl_timer timer = p->cmd_timer;
-        vk->CmdWriteTimestamp(cmd->buf, VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
+        vk->CmdWriteTimestamp(cmd->buf, VK_PIPELINE_STAGE_ALL_COMMANDS_BIT,
                               timer->qpool, timer->index_write + 1);
 
         timer->pending |= timer_bit(timer->index_write);
-        vk_cmd_callback(cmd, (vk_cb) timer_end_cb, timer,
+        vk_cmd_callback(cmd, timer_end_cb, timer,
                         (void *) (uintptr_t) timer->index_write);
 
         timer->index_write = (timer->index_write + 2) % QUERY_POOL_SIZE;
@@ -239,19 +229,32 @@ bool _end_cmd(pl_gpu gpu, struct vk_cmd **pcmd, bool submit)
         vk->CmdEndDebugUtilsLabelEXT(cmd->buf);
 
     if (submit)
-        ret = vk_cmd_submit(vk, &p->cmd);
+        ret = vk_cmd_submit(&p->cmd);
 
     pl_mutex_unlock(&p->recording);
     return ret;
 }
 
+void vk_gpu_idle_callback(pl_gpu gpu, vk_cb cb, const void *priv, const void *arg)
+{
+    struct pl_vk *p = PL_PRIV(gpu);
+    struct vk_ctx *vk = p->vk;
+
+    pl_mutex_lock(&p->recording);
+    if (p->cmd) {
+        vk_cmd_callback(p->cmd, cb, priv, arg);
+    } else {
+        vk_dev_callback(vk, cb, priv, arg);
+    }
+    pl_mutex_unlock(&p->recording);
+}
+
 static void vk_gpu_destroy(pl_gpu gpu)
 {
     struct pl_vk *p = PL_PRIV(gpu);
     struct vk_ctx *vk = p->vk;
 
-    pl_dispatch_destroy(&p->dp);
-    vk_cmd_submit(vk, &p->cmd);
+    vk_cmd_submit(&p->cmd);
     vk_wait_idle(vk);
 
     for (enum pl_tex_sample_mode s = 0; s < PL_TEX_SAMPLE_MODE_COUNT; s++) {
@@ -259,7 +262,7 @@ static void vk_gpu_destroy(pl_gpu gpu)
             vk->DestroySampler(vk->dev, p->samplers[s][a], PL_VK_ALLOC);
     }
 
-    spirv_compiler_destroy(&p->spirv);
+    pl_spirv_destroy(&p->spirv);
     pl_mutex_destroy(&p->recording);
     pl_free((void *) gpu);
 }
@@ -352,7 +355,7 @@ static pl_handle_caps vk_tex_handle_caps(struct vk_ctx *vk, bool import)
         };
 
         VkResult res;
-        res = vk->GetPhysicalDeviceImageFormatProperties2KHR(vk->physd, &pinfo, &props);
+        res = vk->GetPhysicalDeviceImageFormatProperties2(vk->physd, &pinfo, &props);
         if (res != VK_SUCCESS) {
             PL_DEBUG(vk, "Tex caps for %s (0x%x) unsupported: %s",
                      vk_handle_name(ext_pinfo.handleType),
@@ -381,6 +384,35 @@ static const VkFilter filters[PL_TEX_SAMPLE_MODE_COUNT] = {
     [PL_TEX_SAMPLE_LINEAR]  = VK_FILTER_LINEAR,
 };
 
+static inline struct pl_spirv_version get_spirv_version(const struct vk_ctx *vk)
+{
+    if (vk->api_ver >= VK_API_VERSION_1_4) {
+        return (struct pl_spirv_version) {
+            .env_version = VK_API_VERSION_1_4,
+            .spv_version = PL_SPV_VERSION(1, 6),
+        };
+    }
+
+    if (vk->api_ver >= VK_API_VERSION_1_3) {
+        const VkPhysicalDeviceMaintenance4Features *device_maintenance4;
+        device_maintenance4 = vk_find_struct(&vk->features,
+            VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_4_FEATURES);
+
+        if (device_maintenance4 && device_maintenance4->maintenance4) {
+            return (struct pl_spirv_version) {
+                .env_version = VK_API_VERSION_1_3,
+                .spv_version = PL_SPV_VERSION(1, 6),
+            };
+        }
+    }
+
+    pl_assert(vk->api_ver >= VK_API_VERSION_1_2);
+    return (struct pl_spirv_version) {
+        .env_version = VK_API_VERSION_1_2,
+        .spv_version = PL_SPV_VERSION(1, 5),
+    };
+}
+
 static const struct pl_gpu_fns pl_fns_vk;
 
 pl_gpu pl_gpu_create_vk(struct vk_ctx *vk)
@@ -392,10 +424,9 @@ pl_gpu pl_gpu_create_vk(struct vk_ctx *vk)
 
     struct pl_vk *p = PL_PRIV(gpu);
     pl_mutex_init(&p->recording);
-    p->impl = pl_fns_vk;
     p->vk = vk;
-
-    p->spirv = spirv_compiler_create(vk->log);
+    p->impl = pl_fns_vk;
+    p->spirv = pl_spirv_create(vk->log, get_spirv_version(vk));
     if (!p->spirv)
         goto error;
 
@@ -441,23 +472,32 @@ pl_gpu pl_gpu_create_vk(struct vk_ctx *vk)
         if (!strcmp(vk->exts.elem[i], VK_KHR_PORTABILITY_SUBSET_EXTENSION_NAME)) {
             vk_link_struct(&props, &port_props);
             is_portability = true;
+            break;
         }
     }
 #endif
 
+    VkPhysicalDeviceHostImageCopyPropertiesEXT copy_props = {
+        .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_IMAGE_COPY_PROPERTIES_EXT,
+    };
+
+    if (vk->CopyMemoryToImageEXT)
+        vk_link_struct(&props, &copy_props);
+
     vk->GetPhysicalDeviceProperties2(vk->physd, &props);
+    VkPhysicalDeviceLimits limits = props.properties.limits;
 
     // Determine GLSL features and limits
     gpu->glsl = (struct pl_glsl_version) {
         .version = 450,
         .vulkan = true,
         .compute = true,
-        .max_shmem_size = vk->limits.maxComputeSharedMemorySize,
-        .max_group_threads = vk->limits.maxComputeWorkGroupInvocations,
+        .max_shmem_size = limits.maxComputeSharedMemorySize,
+        .max_group_threads = limits.maxComputeWorkGroupInvocations,
         .max_group_size = {
-            vk->limits.maxComputeWorkGroupSize[0],
-            vk->limits.maxComputeWorkGroupSize[1],
-            vk->limits.maxComputeWorkGroupSize[2],
+            limits.maxComputeWorkGroupSize[0],
+            limits.maxComputeWorkGroupSize[1],
+            limits.maxComputeWorkGroupSize[2],
         },
     };
 
@@ -476,49 +516,58 @@ pl_gpu pl_gpu_create_vk(struct vk_ctx *vk)
     }
 
     if (vk->features.features.shaderImageGatherExtended) {
-        gpu->glsl.min_gather_offset = vk->limits.minTexelGatherOffset;
-        gpu->glsl.max_gather_offset = vk->limits.maxTexelGatherOffset;
+        gpu->glsl.min_gather_offset = limits.minTexelGatherOffset;
+        gpu->glsl.max_gather_offset = limits.maxTexelGatherOffset;
     }
 
+    const size_t max_size = vk_malloc_avail(vk->ma, 0);
+    const size_t max_vram = vk_malloc_avail(vk->ma, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);
     gpu->limits = (struct pl_gpu_limits) {
         // pl_gpu
         .thread_safe        = true,
         .callbacks          = true,
         // pl_buf
-        .max_buf_size       = SIZE_MAX, // no limit imposed by vulkan
-        .max_ubo_size       = vk->limits.maxUniformBufferRange,
-        .max_ssbo_size      = vk->limits.maxStorageBufferRange,
-        .max_vbo_size       = SIZE_MAX,
-        .max_mapped_size    = SIZE_MAX,
-        .max_buffer_texels  = vk->limits.maxTexelBufferElements,
+        .max_buf_size       = max_size,
+        .max_ubo_size       = PL_MIN(limits.maxUniformBufferRange, max_size),
+        .max_ssbo_size      = PL_MIN(limits.maxStorageBufferRange, max_size),
+        .max_vbo_size       = max_vram,
+        .max_mapped_size    = vk_malloc_avail(vk->ma, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT),
+        .max_mapped_vram    = vk_malloc_avail(vk->ma, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT |
+                                                      VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT),
+        .max_buffer_texels  = PL_MIN(limits.maxTexelBufferElements, max_size),
         .align_host_ptr     = host_props.minImportedHostPointerAlignment,
+        .host_cached        = vk_malloc_avail(vk->ma, VK_MEMORY_PROPERTY_HOST_CACHED_BIT),
+        .host_ptr_slow      = props.properties.vendorID == VK_VENDOR_ID_NVIDIA,
         // pl_tex
-        .max_tex_1d_dim     = vk->limits.maxImageDimension1D,
-        .max_tex_2d_dim     = vk->limits.maxImageDimension2D,
-        .max_tex_3d_dim     = vk->limits.maxImageDimension3D,
+        .max_tex_1d_dim     = limits.maxImageDimension1D,
+        .max_tex_2d_dim     = limits.maxImageDimension2D,
+        .max_tex_3d_dim     = limits.maxImageDimension3D,
         .blittable_1d_3d    = true,
         .buf_transfer       = true,
-        .align_tex_xfer_pitch  = vk->limits.optimalBufferCopyRowPitchAlignment,
-        .align_tex_xfer_offset = pl_lcm(vk->limits.optimalBufferCopyOffsetAlignment, 4),
+        .align_tex_xfer_pitch  = limits.optimalBufferCopyRowPitchAlignment,
+        .align_tex_xfer_offset = pl_lcm(limits.optimalBufferCopyOffsetAlignment, 4),
         // pl_pass
         .max_variable_comps = 0, // vulkan doesn't support these at all
         .max_constants      = SIZE_MAX,
         .array_size_constants = !is_portability,
-        .max_pushc_size     = vk->limits.maxPushConstantsSize,
+        .max_pushc_size     = limits.maxPushConstantsSize,
 #ifdef VK_KHR_portability_subset
         .align_vertex_stride = port_props.minVertexInputBindingStrideAlignment,
 #else
         .align_vertex_stride = 1,
 #endif
         .max_dispatch = {
-            vk->limits.maxComputeWorkGroupCount[0],
-            vk->limits.maxComputeWorkGroupCount[1],
-            vk->limits.maxComputeWorkGroupCount[2],
+            limits.maxComputeWorkGroupCount[0],
+            limits.maxComputeWorkGroupCount[1],
+            limits.maxComputeWorkGroupCount[2],
         },
         .fragment_queues    = vk->pool_graphics->num_queues,
         .compute_queues     = vk->pool_compute->num_queues,
     };
 
+    // Allow 1 kB of wiggle room in case of some internal reservation or off-by-one
+    p->rebar_enabled = (max_vram - gpu->limits.max_mapped_vram) < 1024;
+
     gpu->export_caps.buf = vk_malloc_handle_caps(vk->ma, false);
     gpu->import_caps.buf = vk_malloc_handle_caps(vk->ma, true);
     gpu->export_caps.tex = vk_tex_handle_caps(vk, false);
@@ -539,12 +588,14 @@ pl_gpu pl_gpu_create_vk(struct vk_ctx *vk)
     if (vk->CmdPushDescriptorSetKHR)
         p->max_push_descriptors = pushd_props.maxPushDescriptors;
 
-    if (vk->ResetQueryPoolEXT) {
-        const VkPhysicalDeviceHostQueryResetFeatures *host_query_reset;
-        host_query_reset = vk_find_struct(&vk->features,
-            VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_QUERY_RESET_FEATURES);
-        if (host_query_reset)
-            p->host_query_reset = host_query_reset->hostQueryReset;
+    if (copy_props.copyDstLayoutCount || copy_props.copySrcLayoutCount) {
+        PL_ARRAY_REALLOC(gpu, p->host_ul_layouts, copy_props.copyDstLayoutCount);
+        PL_ARRAY_REALLOC(gpu, p->host_dl_layouts, copy_props.copySrcLayoutCount);
+        p->host_ul_layouts.num = copy_props.copyDstLayoutCount;
+        p->host_dl_layouts.num = copy_props.copySrcLayoutCount;
+        copy_props.pCopyDstLayouts = p->host_ul_layouts.elem;
+        copy_props.pCopySrcLayouts = p->host_dl_layouts.elem;
+        vk->GetPhysicalDeviceProperties2(vk->physd, &props);
     }
 
     vk_setup_formats(gpu);
@@ -552,7 +603,7 @@ pl_gpu pl_gpu_create_vk(struct vk_ctx *vk)
     // Compute the correct minimum texture alignment
     p->min_texel_alignment = 1;
     for (int i = 0; i < gpu->num_formats; i++) {
-        if (gpu->formats[i]->emulated)
+        if (gpu->formats[i]->emulated || gpu->formats[i]->opaque)
             continue;
         size_t texel_size = gpu->formats[i]->texel_size;
         p->min_texel_alignment = pl_lcm(p->min_texel_alignment, texel_size);
@@ -582,8 +633,6 @@ pl_gpu pl_gpu_create_vk(struct vk_ctx *vk)
         }
     }
 
-    // Create the dispatch last, after any setup of `gpu` is done
-    p->dp = pl_dispatch_create(vk->log, gpu);
     return pl_gpu_finalize(gpu);
 
 error:
@@ -591,74 +640,37 @@ error:
     return NULL;
 }
 
-static void vk_sync_destroy(pl_gpu gpu, pl_sync sync)
+void pl_vulkan_sem_destroy(pl_gpu gpu, VkSemaphore *semaphore)
 {
-    if (!sync)
+    VkSemaphore sem = *semaphore;
+    if (!sem)
         return;
 
     struct pl_vk *p = PL_PRIV(gpu);
     struct vk_ctx *vk = p->vk;
-    struct pl_sync_vk *sync_vk = PL_PRIV(sync);
-
-#ifdef PL_HAVE_UNIX
-    if (sync->handle_type == PL_HANDLE_FD) {
-        if (sync->wait_handle.fd > -1)
-            close(sync->wait_handle.fd);
-        if (sync->signal_handle.fd > -1)
-            close(sync->signal_handle.fd);
-    }
-#endif
-#ifdef PL_HAVE_WIN32
-    if (sync->handle_type == PL_HANDLE_WIN32) {
-        if (sync->wait_handle.handle != NULL)
-            CloseHandle(sync->wait_handle.handle);
-        if (sync->signal_handle.handle != NULL)
-            CloseHandle(sync->signal_handle.handle);
-    }
-    // PL_HANDLE_WIN32_KMT is just an identifier. It doesn't get closed.
-#endif
-
-    vk->DestroySemaphore(vk->dev, sync_vk->wait, PL_VK_ALLOC);
-    vk->DestroySemaphore(vk->dev, sync_vk->signal, PL_VK_ALLOC);
-
-    pl_free((void *) sync);
+    vk->DestroySemaphore(vk->dev, sem, PL_VK_ALLOC);
+    *semaphore = VK_NULL_HANDLE;
 }
 
-void vk_sync_deref(pl_gpu gpu, pl_sync sync)
-{
-    if (!sync)
-        return;
-
-    struct pl_sync_vk *sync_vk = PL_PRIV(sync);
-    if (pl_rc_deref(&sync_vk->rc))
-        vk_sync_destroy(gpu, sync);
-}
-
-static pl_sync vk_sync_create(pl_gpu gpu, enum pl_handle_type handle_type)
+VkSemaphore pl_vulkan_sem_create(pl_gpu gpu, const struct pl_vulkan_sem_params *params)
 {
     struct pl_vk *p = PL_PRIV(gpu);
     struct vk_ctx *vk = p->vk;
 
-    struct pl_sync_t *sync = pl_zalloc_obj(NULL, sync, struct pl_sync_vk);
-    sync->handle_type = handle_type;
+    pl_assert(PL_ISPOT(params->export_handle));
+    if ((params->export_handle & gpu->export_caps.sync) != params->export_handle) {
+        PL_ERR(gpu, "Invalid handle type 0x%"PRIx64" specified for "
+               "`pl_vulkan_sem_create`!", (uint64_t) params->export_handle);
+        return VK_NULL_HANDLE;
+    }
 
-    struct pl_sync_vk *sync_vk = PL_PRIV(sync);
-    pl_rc_init(&sync_vk->rc);
-
-    VkExportSemaphoreCreateInfoKHR einfo = {
-        .sType = VK_STRUCTURE_TYPE_EXPORT_SEMAPHORE_CREATE_INFO_KHR,
-        .handleTypes = vk_sync_handle_type(handle_type),
-    };
-
-    switch (handle_type) {
+    switch (params->export_handle) {
     case PL_HANDLE_FD:
-        sync->wait_handle.fd = -1;
-        sync->signal_handle.fd = -1;
+        params->out_handle->fd = -1;
         break;
     case PL_HANDLE_WIN32:
     case PL_HANDLE_WIN32_KMT:
-        sync->wait_handle.handle = NULL;
-        sync->signal_handle.handle = NULL;
+        params->out_handle->handle = NULL;
         break;
     case PL_HANDLE_DMA_BUF:
     case PL_HANDLE_HOST_PTR:
@@ -667,55 +679,72 @@ static pl_sync vk_sync_create(pl_gpu gpu, enum pl_handle_type handle_type)
         pl_unreachable();
     }
 
-    const VkSemaphoreCreateInfo sinfo = {
-        .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO,
-        .pNext = &einfo,
+    const VkExportSemaphoreCreateInfoKHR einfo = {
+        .sType = VK_STRUCTURE_TYPE_EXPORT_SEMAPHORE_CREATE_INFO_KHR,
+        .handleTypes = vk_sync_handle_type(params->export_handle),
     };
 
-    VK(vk->CreateSemaphore(vk->dev, &sinfo, PL_VK_ALLOC, &sync_vk->wait));
-    VK(vk->CreateSemaphore(vk->dev, &sinfo, PL_VK_ALLOC, &sync_vk->signal));
-    PL_VK_NAME(SEMAPHORE, sync_vk->wait, "sync wait");
-    PL_VK_NAME(SEMAPHORE, sync_vk->signal, "sync signal");
+    const VkSemaphoreTypeCreateInfo stinfo = {
+        .sType = VK_STRUCTURE_TYPE_SEMAPHORE_TYPE_CREATE_INFO,
+        .pNext = params->export_handle ? &einfo : NULL,
+        .semaphoreType = params->type,
+        .initialValue = params->initial_value,
+    };
+
+    const VkSemaphoreCreateInfo sinfo = {
+        .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO,
+        .pNext = &stinfo,
+    };
+
+    VkSemaphore sem = VK_NULL_HANDLE;
+    VK(vk->CreateSemaphore(vk->dev, &sinfo, PL_VK_ALLOC, &sem));
+    PL_VK_NAME(SEMAPHORE, sem, PL_DEF(params->debug_tag, "pl_vulkan_sem"));
 
 #ifdef PL_HAVE_UNIX
-    if (handle_type == PL_HANDLE_FD) {
+    if (params->export_handle == PL_HANDLE_FD) {
         VkSemaphoreGetFdInfoKHR finfo = {
             .sType = VK_STRUCTURE_TYPE_SEMAPHORE_GET_FD_INFO_KHR,
-            .semaphore = sync_vk->wait,
             .handleType = einfo.handleTypes,
+            .semaphore = sem,
         };
 
-        VK(vk->GetSemaphoreFdKHR(vk->dev, &finfo, &sync->wait_handle.fd));
-
-        finfo.semaphore = sync_vk->signal;
-        VK(vk->GetSemaphoreFdKHR(vk->dev, &finfo, &sync->signal_handle.fd));
+        VK(vk->GetSemaphoreFdKHR(vk->dev, &finfo, &params->out_handle->fd));
     }
 #endif
 
 #ifdef PL_HAVE_WIN32
-    if (handle_type == PL_HANDLE_WIN32 ||
-        handle_type == PL_HANDLE_WIN32_KMT)
+    if (params->export_handle == PL_HANDLE_WIN32 ||
+        params->export_handle == PL_HANDLE_WIN32_KMT)
     {
         VkSemaphoreGetWin32HandleInfoKHR handle_info = {
             .sType = VK_STRUCTURE_TYPE_SEMAPHORE_GET_WIN32_HANDLE_INFO_KHR,
-            .semaphore = sync_vk->wait,
             .handleType = einfo.handleTypes,
+            .semaphore = sem,
         };
 
         VK(vk->GetSemaphoreWin32HandleKHR(vk->dev, &handle_info,
-                                          &sync->wait_handle.handle));
-
-        handle_info.semaphore = sync_vk->signal;
-        VK(vk->GetSemaphoreWin32HandleKHR(vk->dev, &handle_info,
-                                          &sync->signal_handle.handle));
+                                          &params->out_handle->handle));
     }
 #endif
 
-    return sync;
+    return sem;
 
 error:
-    vk_sync_destroy(gpu, sync);
-    return NULL;
+#ifdef PL_HAVE_UNIX
+    if (params->export_handle == PL_HANDLE_FD) {
+        if (params->out_handle->fd > -1)
+            close(params->out_handle->fd);
+    }
+#endif
+#ifdef PL_HAVE_WIN32
+    if (params->export_handle == PL_HANDLE_WIN32) {
+        if (params->out_handle->handle != NULL)
+            CloseHandle(params->out_handle->handle);
+    }
+    // PL_HANDLE_WIN32_KMT is just an identifier. It doesn't get closed.
+#endif
+    vk->DestroySemaphore(vk->dev, sem, PL_VK_ALLOC);
+    return VK_NULL_HANDLE;
 }
 
 static void vk_gpu_flush(pl_gpu gpu)
@@ -754,8 +783,8 @@ struct vk_cmd *pl_vk_steal_cmd(pl_gpu gpu)
 
     struct vk_cmdpool *pool = vk->pool_graphics;
     if (!cmd || cmd->pool != pool) {
-        vk_cmd_submit(vk, &cmd);
-        cmd = vk_cmd_begin(vk, pool, NULL);
+        vk_cmd_submit(&cmd);
+        cmd = vk_cmd_begin(pool, NULL);
     }
 
     return cmd;
@@ -778,7 +807,6 @@ static const struct pl_gpu_fns pl_fns_vk = {
     .tex_upload             = vk_tex_upload,
     .tex_download           = vk_tex_download,
     .tex_poll               = vk_tex_poll,
-    .tex_export             = vk_tex_export,
     .buf_create             = vk_buf_create,
     .buf_destroy            = vk_buf_deref,
     .buf_write              = vk_buf_write,
@@ -790,8 +818,6 @@ static const struct pl_gpu_fns pl_fns_vk = {
     .pass_create            = vk_pass_create,
     .pass_destroy           = vk_pass_destroy,
     .pass_run               = vk_pass_run,
-    .sync_create            = vk_sync_create,
-    .sync_destroy           = vk_sync_deref,
     .timer_create           = vk_timer_create,
     .timer_destroy          = vk_timer_destroy,
     .timer_query            = vk_timer_query,
